#! /usr/bin/python3.6

from .reference import create_reference

geometrical_feature_type = [
    'Unknown',
    'Point',
    'Curve',
    'Line',
    'Circle',
    'Surface',
    'Plane',
    'Solid, Volume'
]


class Part:
    """
    .. note::
        CAA V5 Visual Basic help

        The root level object inside a PartDocument object.
        Role: It aggregates all the objects making up the part document.
        It provides many factories and collections. The collections list only the direct children of the part.
        Selection.Search allows to get all objects of one type.

    :param part: CATIA Part COM object.
    """

    def __init__(self, part):

        self.part = part

    @property
    def name(self):
        """

        :return: part name
        """

        return self.part.Name

    @property
    def density(self):
        """
        .. note::
            CAA V5 Visual Basic help

            Returns the part density.
            | Example:
            | The following example displays the density of the part:
            |     Set partRoot = partDoc.Part
            |       MsgBox "The density is " & partRoot.Density

        :return: density
        """

        return self.part.Density

    def get_annotation_sets(self):
        """
        .. note::
            CAA V5 Visual Basic help

            Returns the collection object containing the annotation sets. All the annotation sets
            that are aggregated in the part might be accessed thru that collection.

            | Example:
            | The following example returns in annotationSets the annotation sets of the partRoot part
            | from the partDoc part document:
            |
            | Set partRoot = partDoc.Part
            | Dim annotationSets As AnnotationSets
            | Set annotationSets = partRoot.AnnotationSets

        :return: AnnotationSets COM object
        """

        return self.part.AnnotationSets

    def get_axes_names(self):
        """
        :return: list of axis names
        """

        names = list()

        for axis in self.get_axis_systems():
            name = axis.name
            names.append(name)

        return names

    def get_axis_systems(self):
        """
        This will not return axis systems inside geometrical sets.
        # todo implement selection search feature to get all axis systems inside parts.

        :return:
        """

        catia_axes = self.part.AxisSystems
        axes = list()

        for i in range(0, catia_axes.Count):
            axis = catia_axes.Item(i + 1)
            axes.append(axis)

        return axes

    def get_axis_by_name(self, name):
        """
        :return: Axis System COM object
        """

        for axis_name in self.get_axes_names():
            if name == axis_name:
                return self.part.AxesSystems.Item(name)
        return None

    def get_bodies(self):
        """
        :return: list()
        """

        catia_bodies = self.part.Bodies
        bodies = list()

        for i in range(0, catia_bodies.Count):
            body = catia_bodies.Item(i + 1)
            bodies.append(body)

        return bodies

    def get_bodies_names(self):
        """
        :return: list()
        """

        names = list()

        for body in self.get_bodies():
            name = body.name
            names.append(name)

        return names

    def get_body_by_name(self, name):
        """
        :return: Body COM object
        """

        for body_name in self.get_bodies_names():
            if name == body_name:
                return self.part.Bodies.Item(name)
        return None

    def get_geometric_elements(self):
        """
        ### FROM CAA V5 Visual Basic help ###
        # Returns the collection object containing the part geometrical elements. Only 3D elements are concerned here,
        # 2D elements are managed in sketches. The origin elements are also accessible thru that collection.
        # Example:
        # The following example returns in geomElts the 3D elements of the partRoot part from the partDoc part document:
        #       Set partRoot = partDoc.Part
        #       Set geomElts = partRoot.GeometricElements

        !!! WARNING !!! The items outputted from this don't return the correct geometrical_feature_type.
        As yet, not sure what impact that will have on measuring.

        elem = geometric_elements.Item(5)
        print(elem.name, part.get_geometrical_feature_type(elem))
        returns Point.1 Unknown

        :return: GeometricElements COM object.
        """

        return self.part.GeometricElements

    def get_geometrical_feature_type(self, geometrical_feature):
        """

        !!! WARNING !!! Use with care on items returned from self.get_geometric_elements()
        Reported types are incorrect with the version of CATIA tested with.

        :param geometrical_feature:
        :return:
        """

        reference_object = create_reference(self.part, geometrical_feature)
        feature_type = self.part.HybridShapeFactory.GetGeometricalFeatureType(reference_object)

        return geometrical_feature_type[feature_type]

    def get_hybrid_bodies(self):
        """

        :return: list()
        """

        hybrid_bodies = self.part.HybridBodies
        bodies = list()

        for i in range(0, hybrid_bodies.Count):
            body = hybrid_bodies.Item(i + 1)
            bodies.append(body)

        return bodies

    def get_hybrid_bodies_names(self):
        """

        :return: list
        """

        hybrid_bodies = self.get_hybrid_bodies()
        bodies = list()

        for hybrid_body in hybrid_bodies:
            body = hybrid_body.Name
            bodies.append(body)

        return bodies

    def get_hybrid_body_by_name(self, name):
        """

        :return: HybridBody COM object.
        """

        for hybrid_body_name in self.get_hybrid_bodies_names():
            if name == hybrid_body_name:
                return self.part.HybridBodies.Item(name)
        return None

    @staticmethod
    def get_hybrid_shapes_from_hybrid_body(hybrid_body):
        """

        :return: list()
        """

        shapes = list()
        for i in range(hybrid_body.HybridShapes.Count):
            shapes.append(hybrid_body.HybridShapes.Item(i + 1))

        return shapes

    def create_geometrical_set(self, name):
        """
        Creates a new geometrical set / HybridBody with name.
        :param name: str()
        :return:
        """

        new_geometrical_set = self.part.HybridBodies.Add()
        new_geometrical_set.Name = name

        return new_geometrical_set

    def update(self):
        """
        .. note::
            CAA V5 Visual Basic help

            Sub Update( )
            Updates of the part result with respect to its specifications. Any composing specification that hasn't its
            result up-to-date will recompute it, thus propagating changes to the whole part.
            | Example:
            | The following example update the part:
            | Set partRoot = partDoc.Part
            |   partRoot.Update


        :return:
        """

        self.part.Update()

    def __repr__(self):
        """
        :return: str()
        """
        return f'Part object (name: {self.name})'



#! /user/bin/python3.6

from win32com.client import Dispatch


def create_measurable(spa_workbench, com_reference):
    """

    :param spa_workbench:
    :param com_reference:
    :return:
    """

    return spa_workbench.GetMeasurable(com_reference)


def run_system_service(vba_code, function_name, measurable_items, cat_script_language=0):
    """

    :param vba_code:
    :param function_name:
    :param measurable_items:
    :param cat_script_language:
    :return:
    """
    try:
        run = Dispatch('CATIA.Application').SystemService.Evaluate(
            vba_code,
            cat_script_language,
            function_name,
            measurable_items,
        )

        return run

    except ValueError:
        # noinspection SpellCheckingInspection
        print(f'There was a problem running SystemService.Evalue(*args) on the VBA code. The inputs were:')
        print(f'vba_code: {vba_code}, function_name: {function_name}')



#! /usr/bin/python3.6


def create_reference(com_part, catia_object):
    """

    :param com_part:
    :param catia_object:
    :return:
    """
    return com_part.CreateReferenceFromObject(catia_object)



#! /usr/bin/python3.6

from win32com.client import Dispatch

from .document import Documents, Document


class CATIAApplication:
    """
    All pycatia scripts will start with the creation on the CATIAApplication() object.

    Usage::

        from pycatia import CATIAApplication
        catia = CATIAApplication()
        documents = catia.documents()
        file_name = 'my_new_part.CATPart'
        documents.open(file_name)

    .. note::
        CAA V5 Visual Basic help

        Represents the current CNext application and its frame window.
        The application is the root object for all the other objects you can use and access from scripts.
        It directly aggregates:

        - The document collection represented by the Documents object. This collection contains all the documents
          currently opened by the application.

        - The window collection represented by the Windows object. This collection contains all the windows currently
          opened by the application, each window displaying one of the documents contained in the document collection

        - The SystemService object, providing information about the system environment.


        The active document and the active window are two key objects for the application you can access using the
        ActiveDocument and ActiveWindow properties respectively. The active window is the window the end user is
        currently working in, and the active document is the document displayed in this active window and that the end
        user is being editing. This document sets its workshop, that is the available menus and toolbars that make it
        possible to edit it, according to its type.

        When you create or use macros for in-process access, the application is always referred to as CATIA.

    """

    def __init__(self):
        self.catia = Dispatch('CATIA.Application')

    def documents(self):
        """
        :return: :class:`Documents()`
        """

        return Documents(self.catia)

    def document(self):
        """
        :return: :class:`Document()`
        """

        return Document(self.catia)

    def refresh_display(self, state=True):
        """
        Set refresh_display to False to speed up macros.

        Set back to True on macro completion.

        .. note::
            CAA V5 Visual Basic help

            Property RefreshDisplay( ) As boolean

            Enables or disables the update of the display during the script replay. To improve performance, this update
            can be temporarily disabled by setting this property to False in the script.
            True (value set by default) if the application's display is refreshed after each method call executed in
            late binding mode . This property does not affect early binding calls nor the get methods because they are
            never followed by a refresh of the display.
            | Example:
            | This example makes the update of the CATIA application's display disabled during the script replay.
            |    CATIA.RefreshDisplay = False

        :return: boolean
        """

        self.catia.RefreshDisplay = state

        return self.catia.RefreshDisplay

    def visible(self, state=True):
        """
        .. note::
            Property Visible( ) As boolean

            Returns or sets the application's window visibility.
            True if the application's window is visible to the end user.
            | Example:
            | This example makes the CATIA application's window visible.
            |   CATIA.Visibility = True

        :param state:
        :return:
        """

        self.catia.Visible = state

        return self.catia.Visible

    def evaluate(self, vba_code, function_name, measurable_items, cat_script_language=0):
        """

        .. note::
            CATIA V5 Visual Basic help

            | Func Evaluate( CATBSTR  iScriptText,
            |       CATScriptLanguage  iLanguage,
            |       CATBSTR  iFunctionName,
            |       CATSafeArrayVariant  iParameters) As CATVariant
            |
            | Evaluates a scripted function.
            | Parameters:
            |   iScriptText
            | The program text
            |   iLanguage
            | The language the program is written in
            |   iFunctionName
            | The name of the function to invoke
            |   iParameters
            | An array of parameters for the function
            |   Result
            | The value returned by the function (if any)
            | Example:
            | This example executes the function CATMain from the CodeToEvaluate string
            |   Dim params()
            |   Dim codeToEvaluate
            |   CodeToEvaluate = "Sub CATMain()" & vbNewLine & _
            |                   "MsgBox " & chr(34) & "Hello World" & chr(34) & vbNewLine & _
            |                   "End Sub"
            |   CATIA.SystemService.Evaluate CodeToEvaluate, CATVBScriptLanguage, "CATMain", params


        :param str() vba_code: String containing script to run.
        :param str() function_name: Name of function within vba script.
        :param list() measurable_items: list of items to pass to script.
        :param int() cat_script_language: An integer representing the language selection from list()
         CATScriptLanguage[CATVBScriptLanguage, CATVBALanguage, CATBasicScriptLanguage, CATJavaLanguage,
         CATJScriptLanguage]
        :return:
        """
        try:
            run = self.catia.SystemService.Evaluate(
                vba_code,
                cat_script_language,
                function_name,
                measurable_items,
            )

            return run

        except ValueError:
            print(f'There was a problem running SystemService.Evaluate(*args) on the VBA code. The inputs were:')
            print(f'vba_code: {vba_code}, function_name: {function_name}')

    def execute_script(self, library_name, library_type, program_name, function_name, items):
        """

        .. note::
            CAA V5 Visual Basic help

            | Func ExecuteScript( CATBSTR  iLibraryName,
            |   CatScriptLibraryType  iType,
            |   CATBSTR  iProgramName,
            |   CATBSTR  iFunctionName,
            |   CATSafeArrayVariant  iParameters) As CATVariant
            |
            | Executes a scripted function.
            | Parameters:
            |   iLibraryName
            |       The library in which the script is contained
            |   iLibraryType
            |       The type of the library
            |   iProgramName
            |       The name of the program in the library
            |   iFunctionName
            |       The name of the function to invoke
            |   iParameters
            |       An array of parameters for the function
            |   Result
            |       The value returned by the function (if any)
            |
            | Example:
            | This example executes the function CATMain in the program Macro1.catvbs contained by Part1.CATPart
            |   Dim params()
            |   CATIA.SystemService.ExecuteScript"Part1.CATPart",
            |       catScriptLibraryTypeDocument,
            |       "Macro1.catvbs",
            |       "CATMain",
            |       params

        :param str() library_name: Full path to location of catia script.
        :param int() library_type:  An integer representing the library type selection from list()
         CatScriptLibraryType[catScriptLibraryTypeDocument, catScriptLibraryTypeDirectory,
         catScriptLibraryTypeVBAProject]
        :param str() program_name: file name of script.
        :param function_name: Name of function to call within script.
        :param list() items: List of items to pass to script.
        :return:
        """

        run = self.catia.SystemService.ExecuteScript(
            library_name,
            library_type,
            program_name,
            function_name,
            items
        )

        return run

    def __repr__(self):
        return '<CATIAApplication object ()>'



#! /usr/bin/python3.6

import os


class HybridShapeFactory:

    def __init__(self, part):
        """

        :param part: CATIA Part COM object.
        """

        self.hsf = part.part.HybridShapeFactory

    # def add_new_point_coord(self, catia, geometrical_set, new_point, point_name):
    #     """
    #     .. note::
    #         CAA V5 Visual Basic help
    #
    #         | Func AddNewPointCoord(double  iX,
    #         |                       double  iY,
    #         |                       double  iZ) As HybridShapePointCoord
    #
    #         | Creates a new point defined by its cartesian coordinates within the current body.
    #         | Parameters:
    #         | iX
    #         |     X coordinate for the point
    #         | iY
    #         |     Y coordinate for the point
    #         | iZ
    #         |     Z coordinate for the point
    #         | oPoint
    #         |     Created point
    #
    #     :param catia: CATIAApplication()
    #     :param geometrical_set:
    #     :param new_point:
    #     :param point_name:
    #     """
    #
    #     point = self.hsf.AddNewPointCoord(new_point[0], new_point[1], new_point[2])
    #
    #     vba_function_name = 'add_point'
    #     vba_function = 'AppendHybridShape'
    #     vba_code = (f'Public Function {vba_function_name}(geometrical_set, point, point_name)\n'
    #                 f'    geometrical_set.{vba_function} point\n'
    #                 f'    point.Name = point_name\n'
    #                 f'End Function')
    #     catia.run_system_service(vba_code, vba_function_name, [geometrical_set, point, point_name])

    def add_new_point_coord(self, catia, geometrical_set, new_point, point_name):
        """

        Replaced run_system service in attempt to solve this issue:
        http://www-01.ibm.com/support/docview.wss?uid=swg1HD79859
        Issue was not resolved. Adding a large number of points gets slower and slower when adding large numbers of
        points. This is visible when adding over 1500 points on test machine (run example_5.py).
        In these cases it's probably best to split the csv file into several files and run the import in batches.

        .. note::
            CAA V5 Visual Basic help

            | Func AddNewPointCoord(double  iX,
            |                       double  iY,
            |                       double  iZ) As HybridShapePointCoord

            | Creates a new point defined by its cartesian coordinates within the current body.
            | Parameters:
            | iX
            |     X coordinate for the point
            | iY
            |     Y coordinate for the point
            | iZ
            |     Z coordinate for the point
            | oPoint
            |     Created point

        :param catia: CATIAApplication()
        :param geometrical_set:
        :param new_point:
        :param point_name:
        """

        macro_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), r'macros')

        point = self.hsf.AddNewPointCoord(new_point[0], new_point[1], new_point[2])

        vba_function_name = 'add_point'

        catia.execute_script(macro_path,
                             1,
                             'pycatia_macros.catvbs',
                             vba_function_name,
                             [geometrical_set, point, point_name])



#! /usr/bin/python3.6

import os

from .catia_application import CATIAApplication
from .csv_tools import create_points
from .csv_tools import csv_reader
from .document import Document
from .document import Documents
from .general_functions import create_measurable
from .general_functions import run_system_service
from .hybridshapefactory import HybridShapeFactory
from .measurable import CATIAMeasurable
from .part import Part
from .product import Product
from .reference import create_reference
from .workbenches import create_spa_workbench


__author__ = 'Paul Bourne'
__author_email = 'evereux@gmail.com'
__description__ = 'A python module to access the CATIA Measurable object.'
__name__ = "pycatia"
__version__ = "0.0.3"
__url__ = "https://github.com/evereux/pycatia"

name = __name__

macro_path = os.path.join(os.path.realpath(__file__), r'macros')
